@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2021
-;; Version: 2.2
+;; Version: 2.3
 ;; Package-Requires: ((emacs "29.1") (compat "30"))
 ;; URL: https://github.com/minad/cape
 ;; Keywords: abbrev, convenience, matching, completion, text
@@ -289,10 +289,10 @@ NAME is the name of the Capf, BEG and END are the input markers."
          (+ beg 0) (+ end 0) (buffer-substring-no-properties beg end)
          str completion-ignore-case
          (if completion-regexp-list
-             (format " regexp=%s" (cape--debug-print completion-regexp-list t))
+             (concat " regexp=" (cape--debug-print completion-regexp-list t))
            "")
          (if pred
-             (format " predicate=%s" (cape--debug-print pred))
+             (concat " predicate=" (cape--debug-print pred))
            "")
          (cape--debug-print result)))
       result)))
@@ -582,21 +582,22 @@ If INTERACTIVE is nil the function acts like a Capf."
     (require 'dabbrev))
   (let ((re (or dabbrev-abbrev-char-regexp "\\sw\\|\\s_"))
         (limit (minibuffer-prompt-end)))
-    (when (or (looking-at re)
-              (and (> (point) limit)
-                   (save-excursion (forward-char -1) (looking-at re))))
-      (cons (save-excursion
-              (while (and (> (point) limit)
-                          (save-excursion (forward-char -1) (looking-at re)))
-                (forward-char -1))
-              (when dabbrev-abbrev-skip-leading-regexp
-                (while (looking-at dabbrev-abbrev-skip-leading-regexp)
-                  (forward-char 1)))
-              (point))
-            (save-excursion
-              (while (looking-at re)
-                (forward-char 1))
-              (point))))))
+    (if (or (looking-at re)
+            (and (> (point) limit)
+                 (save-excursion (forward-char -1) (looking-at re))))
+        (cons (save-excursion
+                (while (and (> (point) limit)
+                            (save-excursion (forward-char -1) (looking-at re)))
+                  (forward-char -1))
+                (when dabbrev-abbrev-skip-leading-regexp
+                  (while (looking-at dabbrev-abbrev-skip-leading-regexp)
+                    (forward-char 1)))
+                (point))
+              (save-excursion
+                (while (looking-at re)
+                  (forward-char 1))
+                (point)))
+      (cons (point) (point)))))
 
 ;;;###autoload
 (defun cape-dabbrev (&optional interactive)
@@ -609,10 +610,10 @@ See the user option `cape-dabbrev-buffer-function'."
   (interactive (list t))
   (if interactive
       (cape-interactive #'cape-dabbrev)
-    (when-let ((bounds (cape--dabbrev-bounds)))
-      `(,(car bounds) ,(cdr bounds)
+    (pcase-let ((`(,beg . ,end) (cape--dabbrev-bounds)))
+      `(,beg ,end
         ,(completion-table-case-fold
-          (cape--dynamic-table (car bounds) (cdr bounds) #'cape--dabbrev-list)
+          (cape--dynamic-table beg end #'cape--dabbrev-list)
           (not (cape--case-fold-p dabbrev-case-fold-search)))
         ,@cape--dabbrev-properties))))
 
@@ -882,6 +883,12 @@ again if the input prefix changed."
     (interactive (list t))
     (if interactive (cape-interactive capf) (funcall capf))))
 
+(defvar cape--super-functions
+  '( :company-docsig :company-location :company-kind
+     :company-doc-buffer :company-deprecated
+     :annotation-function :exit-function)
+  "List of extra functions which are handled by `cape-wrap-super'.")
+
 ;;;###autoload
 (defun cape-wrap-super (&rest capfs)
   "Call CAPFS and return merged completion result.
@@ -892,10 +899,13 @@ result.  Such behavior is useful when listing multiple super Capfs in
 the `completion-at-point-functions':
 
   (setq completion-at-point-functions
-        (list (cape-capf-super \\='eglot-completion-at-point
+        (list (cape-capf-super \\='elisp-completion-at-point
                                :with \\='tempel-complete)
               (cape-capf-super \\='cape-dabbrev
-                               :with \\='tempel-complete)))"
+                               :with \\='tempel-complete)))
+
+See the dual `cape-wrap-choose' if you want to try multiple Capfs in
+turn."
   (when-let ((results (cl-loop for capf in capfs until (eq capf :with)
                                for res = (funcall capf)
                                if res collect (cons t res))))
@@ -907,11 +917,7 @@ the `completion-at-point-functions':
                  (cand-ht nil)
                  (tables nil)
                  (exclusive nil)
-                 (prefix-len nil)
-                 (cand-functions
-                  '( :company-docsig :company-location :company-kind
-                     :company-doc-buffer :company-deprecated
-                     :annotation-function :exit-function)))
+                 (prefix-len nil))
       (cl-loop for (main beg2 end2 table . plist) in results do
                ;; Note: `cape-capf-super' currently cannot merge Capfs which
                ;; trigger at different beginning positions.  In order to support
@@ -924,7 +930,7 @@ the `completion-at-point-functions':
                              (mapcan (lambda (f)
                                        (when-let ((v (plist-get plist f)))
                                          (list f v)))
-                                     cand-functions))
+                                     cape--super-functions))
                        tables)
                  ;; The resulting merged Capf is exclusive if one of the main
                  ;; Capfs is exclusive.
@@ -987,10 +993,10 @@ the `completion-at-point-functions':
                          (if (and table-pred pred)
                              (lambda (x) (and (funcall table-pred x) (funcall pred x)))
                            (or table-pred pred)))))))
-         :company-prefix-length ,prefix-len
          :category cape-super
-         :display-sort-function identity
-         :cycle-sort-function identity
+         :company-prefix-length ,prefix-len
+         :display-sort-function ,#'identity
+         :cycle-sort-function ,#'identity
          ,@(and (not exclusive) '(:exclusive no))
          ,@(mapcan
             (lambda (prop)
@@ -1002,7 +1008,28 @@ the `completion-at-point-functions':
                         (when-let ((plist (and cand-ht (gethash cand cand-ht)))
                                    (fun (plist-get plist prop)))
                           (apply fun cand args))))))
-            cand-functions)))))
+            cape--super-functions)))))
+
+;;;###autoload
+(defun cape-wrap-choose (&rest capfs)
+  "Call each of CAPFS in turn and return first non-nil result.
+Use `cape-wrap-choose' to create a single Capf from multiple Capfs.
+Usually you want to add multiple non-exclusive Capfs to the variable
+`completion-at-point-functions' directly instead.  See the dual
+`cape-wrap-super' if you want to merge multiple Capf results."
+  (cl-loop
+   for capf in capfs thereis
+   (pcase (funcall capf)
+     ((and result `(,beg ,end ,table . ,plist))
+      (let* ((str (buffer-substring-no-properties beg end))
+             (pt (- (point) beg))
+             (pred (plist-get plist :predicate))
+             (md (completion-metadata (substring str 0 pt) table pred)))
+        ;; NOTE: Treat the Capfs always as non-exclusive. Return the first which
+        ;; returns a non-nil result. See the comment in `corfu--capf-wrapper'
+        ;; for further considerations.
+        (and (completion-try-completion str table pred pt md)
+             result))))))
 
 ;;;###autoload
 (defun cape-wrap-debug (capf &optional name)
@@ -1040,10 +1067,10 @@ meaningful debugging output."
           table name (copy-marker beg) (copy-marker end t))
         ,@(when-let ((exit (plist-get plist :exit-function)))
             (list :exit-function
-                  (lambda (cand status)
-                    (cape--debug-message "%s:exit(candidate=%S status=%s)"
-                                         name cand status)
-                    (funcall exit cand status))))
+                  (lambda (str status)
+                    (cape--debug-message "%s:exit(status=%s string=%S)"
+                                         name status str)
+                    (funcall exit str status))))
         . ,plist))
     (result
      (cape--debug-message "%s() => %s (No completion)"
@@ -1086,7 +1113,8 @@ completion table is refreshed on every input change."
 
 ;;;###autoload
 (defun cape-wrap-passthrough (capf)
-  "Call CAPF and make sure that no completion style filtering takes place."
+  "Call CAPF and make sure that no completion style filtering takes place.
+This function can be used as an advice around an existing Capf."
   (pcase (funcall capf)
     (`(,beg ,end ,table . ,plist)
      `(,beg ,end ,(cape--passthrough-table table) ,@plist))))
@@ -1096,12 +1124,11 @@ completion table is refreshed on every input change."
   "Call CAPF and strip or add completion PROPERTIES.
 Completion properties include for example :exclusive, :category,
 :annotation-function, :display-sort-function and various :company-*
-extensions.  The :strip flag means to strip all completion properties."
+extensions.  Strip all properties if PROPERTIES is :strip."
   (pcase (funcall capf)
     (`(,beg ,end ,table . ,plist)
      `( ,beg ,end ,table
-        ,@(and (not (plist-get properties :strip))
-               (append properties plist))))))
+        ,@(and (not (eq :strip (car properties))) (append properties plist))))))
 
 ;;;###autoload
 (defun cape-wrap-nonexclusive (capf)
@@ -1110,9 +1137,11 @@ This function can be used as an advice around an existing Capf."
   (cape-wrap-properties capf :exclusive 'no))
 
 ;;;###autoload
-(defun cape-wrap-sort (capf sort)
-  "Call CAPF and add SORT function.
-This function can be used as an advice around an existing Capf."
+(defun cape-wrap-sort (capf &optional sort)
+  "Call CAPF and add SORT function as completion metadata.
+If the SORT argument is nil or not given, the completion UI will use
+its own default sorting algorithm.  This function can be used as an
+advice around an existing Capf."
   (cape-wrap-properties
    capf
    :display-sort-function sort
@@ -1176,8 +1205,7 @@ If the prefix is long enough, enforce auto completion."
 
 ;;;###autoload
 (defun cape-wrap-inside-faces (capf &rest faces)
-  "Call CAPF only if inside FACES.
-This function can be used as an advice around an existing Capf."
+  "Call CAPF only if inside FACES."
   (when-let (((> (point) (point-min)))
              (fs (get-text-property (1- (point)) 'face))
              ((if (listp fs)
@@ -1204,26 +1232,6 @@ This function can be used as an advice around an existing Capf."
 This function can be used as an advice around an existing Capf."
   (and (nth 3 (syntax-ppss)) (funcall capf)))
 
-;;;###autoload
-(defun cape-wrap-purify (capf)
-  "Call CAPF and ensure that it does not illegally modify the buffer.
-This function can be used as an advice around an existing Capf."
-  ;; bug#50470: Fix Capfs which illegally modify the buffer or which illegally
-  ;; call `completion-in-region'.  The workaround here was proposed by
-  ;; @jakanakaevangeli and is used in his capf-autosuggest package.  In Emacs 29
-  ;; the purity bug of Pcomplete has been fixed, such that make
-  ;; `cape-wrap-purify' is not necessary anymore.
-  (catch 'cape--illegal-completion-in-region
-    (condition-case nil
-        (let ((buffer-read-only t)
-              (inhibit-read-only nil)
-              (completion-in-region-function
-               (lambda (beg end coll pred)
-                 (throw 'cape--illegal-completion-in-region
-                        (list beg end coll :predicate pred)))))
-          (funcall capf))
-      (buffer-read-only nil))))
-
 ;;;###autoload
 (defun cape-wrap-accept-all (capf)
   "Call CAPF and return a completion table which accepts every input.
@@ -1232,9 +1240,69 @@ This function can be used as an advice around an existing Capf."
     (`(,beg ,end ,table . ,plist)
      `(,beg ,end ,(cape--accept-all-table table) . ,plist))))
 
+(defvar cape--trigger-syntax-table (make-syntax-table (syntax-table))
+  "Syntax table used for the trigger character.")
+
+;;;###autoload
+(defun cape-wrap-trigger (capf trigger)
+  "Ensure that TRIGGER character occurs before point and then call CAPF.
+See also `corfu-auto-trigger'.
+Example:
+  (setq corfu-auto-trigger \"/\"
+        completion-at-point-functions
+        (list (cape-capf-trigger \\='cape-abbrev ?/)))"
+  (when-let ((pos (save-excursion (search-backward (char-to-string trigger) (pos-bol) 'noerror)))
+             ((save-excursion (not (re-search-backward "\\s-" pos 'noerror)))))
+    (pcase
+        ;; Treat the trigger character as punctuation.
+        (with-syntax-table cape--trigger-syntax-table
+          (unless (eq (char-syntax trigger) ?.)
+            (modify-syntax-entry trigger "."))
+          (funcall capf))
+      (`(,beg ,end ,table . ,plist)
+       (when (<= pos beg (1+ pos))
+         `( ,(1+ pos) ,end ,table
+            :company-prefix-length t
+            :exit-function
+            ,(let ((pos (copy-marker pos))
+                   (end (copy-marker (1+ pos))))
+               (lambda (str status)
+                 (delete-region pos end)
+                 (when-let ((exit (plist-get plist :exit-function)))
+                   (funcall exit str status))))
+            . ,plist))))))
+
+;;;###autoload (autoload 'cape-capf-purify "cape")
+;;;###autoload
+(defun cape-wrap-purify (capf)
+  "Obsolete purification wrapper calling CAPF.
+This function can be used as an advice around an existing Capf."
+  (warn "`cape-wrap-purify' is obsolete")
+  (funcall capf))
+(make-obsolete 'cape-wrap-purify nil "2.2")
+(make-obsolete 'cape-capf-purify nil "2.2")
+
+(dolist (wrapper (list #'cape-wrap-accept-all #'cape-wrap-buster
+                       #'cape-wrap-case-fold #'cape-wrap-choose
+                       #'cape-wrap-debug #'cape-wrap-inside-code
+                       #'cape-wrap-inside-comment #'cape-wrap-inside-faces
+                       #'cape-wrap-inside-string #'cape-wrap-nonexclusive
+                       #'cape-wrap-noninterruptible #'cape-wrap-passthrough
+                       #'cape-wrap-predicate #'cape-wrap-prefix-length
+                       #'cape-wrap-properties 'cape-wrap-purify
+                       #'cape-wrap-silent #'cape-wrap-sort
+                       #'cape-wrap-super #'cape-wrap-trigger))
+  (let ((name (string-remove-prefix "cape-wrap-" (symbol-name wrapper))))
+    (defalias (intern (format "cape-capf-%s" name))
+      (lambda (capf &rest args) (lambda () (apply wrapper capf args)))
+      (format "Create a %s Capf from CAPF.
+The Capf calls `%s' with CAPF and ARGS as arguments.
+See `%s' for documentation." name wrapper wrapper))))
+
 ;;;###autoload (autoload 'cape-capf-accept-all "cape")
 ;;;###autoload (autoload 'cape-capf-buster "cape")
 ;;;###autoload (autoload 'cape-capf-case-fold "cape")
+;;;###autoload (autoload 'cape-capf-choose "cape")
 ;;;###autoload (autoload 'cape-capf-debug "cape")
 ;;;###autoload (autoload 'cape-capf-inside-code "cape")
 ;;;###autoload (autoload 'cape-capf-inside-comment "cape")
@@ -1246,24 +1314,9 @@ This function can be used as an advice around an existing Capf."
 ;;;###autoload (autoload 'cape-capf-predicate "cape")
 ;;;###autoload (autoload 'cape-capf-prefix-length "cape")
 ;;;###autoload (autoload 'cape-capf-properties "cape")
-;;;###autoload (autoload 'cape-capf-purify "cape")
 ;;;###autoload (autoload 'cape-capf-silent "cape")
 ;;;###autoload (autoload 'cape-capf-super "cape")
-
-(dolist (wrapper (list #'cape-wrap-accept-all #'cape-wrap-buster
-                       #'cape-wrap-case-fold #'cape-wrap-debug
-                       #'cape-wrap-inside-code #'cape-wrap-inside-comment
-                       #'cape-wrap-inside-faces #'cape-wrap-inside-string
-                       #'cape-wrap-nonexclusive #'cape-wrap-noninterruptible
-                       #'cape-wrap-passthrough #'cape-wrap-predicate
-                       #'cape-wrap-prefix-length #'cape-wrap-properties
-                       #'cape-wrap-purify #'cape-wrap-silent
-                       #'cape-wrap-sort #'cape-wrap-super))
-  (let ((name (string-remove-prefix "cape-wrap-" (symbol-name wrapper))))
-    (defalias (intern (format "cape-capf-%s" name))
-      (lambda (capf &rest args) (lambda () (apply wrapper capf args)))
-      (format "Create a %s Capf from CAPF.
-The Capf calls `%s' with CAPF and ARGS as arguments." name wrapper))))
+;;;###autoload (autoload 'cape-capf-trigger "cape")
 
 (defvar-keymap cape-prefix-map
   :doc "Keymap used as completion entry point.
